@@ -27,12 +27,12 @@ class directory_backend:
     fd.close()
     self._nimbstor.append_block_info(number, checksum1, checksum2, self)
 
-  def read_buffer(self, number, checksum1, checksum2, decode_func):
+  def read_buffer(self, number, checksum1, checksum2):
     fname = os.path.join(self._directory, "%07d.%s.%s" % (number, checksum1, checksum2))
     fd = open(fname, 'r')
     buffer = fd.read()
     fd.close()
-    return decode_func(buffer)
+    yield buffer
 
   def close(self):
     pass
