@@ -2,11 +2,69 @@
 
 BLOCK_SIZE = 1048576 * 4
 
-import os, sys, hashlib, lzma, aes, time, zlib, bz2, email, types
+import os, sys, hashlib, lzma, aes, time, zlib, bz2, timeit
 import cPickle as pickle
 import adler64
+import multiprocessing
 from itertools import ifilter
 
+def message_encode(buffer, password, meta, compression, bufnum, checksum1, checksum2):
+  """Internal function to encode message.
+
+  Args:
+    buffer (str): Data buffer.
+    password (str): The password as SHA256 hash value.
+    meta (boolean): Meta blocks are handled somewhat different.
+    compression (tuple): Compression type to use.
+    bufnum (int): Current buffer number
+    checksum1 (int): Adler64 checksum.
+    checksum2 (str): Checksum as returned by stor_stream.checksum(buffer)
+  """
+  if meta: buffer = bz2.compress(buffer, 9)
+  elif compression == ('lzma', 10): buffer = lzma.compress(buffer, {'level': 9, 'extreme': True})
+  elif compression[0] == 'lzma': buffer = lzma.compress(buffer, {'level': compression[1]})
+  elif compression[0] == 'gzip': buffer = zlib.compress(buffer, compression[1])
+  elif compression[0] == 'bzip2': bffer = bz2.compress(buffer, compression[1])
+  elif compression[0] != None:
+    raise SystemError("Unsupported compression type %s" % repr(compression))
+  bufrest = len(buffer) % 16
+  if bufrest > 0: buffer = buffer + "\0" * (16 - bufrest)
+  if password: buffer = aes.Keysetup(password).encrypt(buffer)
+  return (buffer, bufnum, checksum1, checksum2)
+
+def message_decode(buffer, password, cksumpwd, meta, compression, bufnum, checksum1, checksum2, bufflen):
+  """Internal function to decode message.
+
+  Args:
+    buffer (str): Data buffer.
+    password (str): The password as SHA256 hash value.
+    cksumpwd (str): The password for checksum (MD5 hash value).
+    meta (boolean): Meta blocks are handled somewhat different.
+    compression (tuple): Compression type to use.
+    bufnum (int): Current buffer number
+    checksum1 (int): Adler64 checksum.
+    checksum2 (str): Checksum as returned by stor_stream.checksum(buffer)
+    bufflen (str): Correct data length to return or 0.
+  """
+  try:
+    if password: buffer = aes.Keysetup(password).decrypt(buffer)
+    if meta: buffer = bz2.decompress(buffer)
+    elif compression[0] == 'lzma': buffer = lzma.decompress(buffer)
+    elif compression[0] == 'gzip': buffer = zlib.decompress(buffer)
+    elif compression[0] == 'bzip2': buffer = bz2.decompress(buffer)
+    elif compression[0] != None:
+      raise SystemError("Unsupported compression type %s" % repr(compression))
+    if bufflen != 0 and len(buffer) > bufflen: buffer = buffer[:bufflen]
+    ctx = hashlib.sha256()
+    if cksumpwd != None: ctx.update(cksumpwd)
+    ctx.update(buffer)
+    newcksum = ctx.hexdigest().upper()
+    if newcksum != checksum2: raise SystemError()
+  except Exception, err:
+    raise SystemError("Wrong password or invalid data (%s)" % repr(err))
+  return (buffer, bufnum, checksum1, checksum2, bufflen)
+
+
 class stor_buffer:
   """Manage buffer for nimbstor.
 
@@ -14,7 +72,7 @@ class stor_buffer:
 
   """
 
-  def __init__(self, data = ""):
+  def __init__(self, data = "", splitsize = None):
     """Initialize the buffer.
 
     Args:
@@ -26,6 +84,7 @@ class stor_buffer:
     self._data = [data]
     self._len = len(data)
     self._val = None
+    self._splitsize = splitsize
 
   def __len__(self):
     """Return buffer content length.
@@ -49,8 +108,17 @@ class stor_buffer:
     'zxcvasdfqwer'
     """
     self._len += len(data)
-    if left: self._data.insert(0, data)
-    else: self._data.append(data)
+    if self._splitsize != None:
+      listdat = []
+      curpos = 0
+      while curpos < self._len:
+        listdat.append(data[curpos:curpos + self._splitsize])
+        curpos += self._splitsize
+      if left: self._data[0:0] = listdat
+      else: self._data.extend(listdat)
+    else:
+      if left: self._data.insert(0, data)
+      else: self._data.append(data)
     self._val = None
 
   def _rmsize(self, size):
@@ -85,11 +153,31 @@ class stor_buffer:
     >>> buf.value()
     'fqwer'
     """
-    if self._val == None: self._val = "".join(self._data)
-    if size == None: size = self._len
-    ret = self._val[:size]
-    self._rmsize(size)
-    return ret
+    if size == None:
+      buf = "".join(self._data)
+      self._data = []
+      self._len = 0
+      self._val = None
+      return buf
+    if len(self._data) == 0:
+      return ''
+    self._val = None
+    buf = []
+    restsize = size
+    while restsize > 0 and len(self._data) > 0:
+      curbuf = self._data[0]
+      curlen = len(curbuf)
+      if curlen > restsize:
+        buf.append(curbuf[:restsize])
+        self._data[0] = curbuf[restsize:]
+        self._len -= restsize
+        restsize = 0
+      else:
+        buf.append(curbuf)
+        del self._data[0]
+        self._len -= curlen
+        restsize -= curlen
+    return ''.join(buf)
 
   def truncate(self, size = None):
     """Remove some amount of data from buffer.
@@ -135,8 +223,9 @@ class stor_stream:
     self._known_blocks = {}
     self._known_parts = {}
     self._known_metadata = []
+    self._backend = None
     if password != None:
-      self._aes = aes.Keysetup(hashlib.sha256(password).digest())
+      self._aes = hashlib.sha256(password).digest()
       self._pass = hashlib.md5(password).digest()
     else:
       self._aes = None
@@ -151,6 +240,14 @@ class stor_stream:
     """
     self._backend = backend
 
+  def read_buffer(self, number, checksum1, checksum2):
+    """Forward read_buffer to the backend."""
+    return self._backend.read_buffer(number, checksum1, checksum2)
+
+  def write_buffer(self, buffer, number, checksum1, checksum2):
+    """Forward write_buffer to the backend."""
+    self._backend.write_buffer(buffer, number, checksum1, checksum2)
+
   def checksum(self, data):
     """Calculate checksum for given data.
     
@@ -199,30 +296,10 @@ class stor_stream:
     """
     self._backend.close()
 
-  def _decode_message(self, buffer, checksum, bufflen, meta = False):
-    """Internal function to decode message.
-
-    Args:
-      buffer (str): Data buffer.
-      checksum (str): Checksum as returned by self.checksum(data)
-      bufflen (str): Correct data length to return or 0.
-      meta (boolean): Meta blocks are handled somewhat different.
-    """
-    data = buffer
-    try:
-      if self._aes: data = self._aes.decrypt(data)
-      if meta: data = bz2.decompress(data)
-      else: data = self._decompress(data)
-      if bufflen != 0 and len(data) > bufflen: data = data[:bufflen]
-      if self.checksum(data) != checksum:
-	raise SystemError()
-    except: raise SystemError("Wrong password or invalid data")
-    return data
-
 class stor_output(stor_stream):
   """Storage output stream."""
 
-  def __init__(self, password, compression, parent, description, keywords, onlyifnew = False):
+  def __init__(self, password, compression, parent, description, keywords, parlimit = 4, onlyifnew = False):
     """Initialization of output stream.
 
     Args:
@@ -244,19 +321,18 @@ class stor_output(stor_stream):
 	'blocks': [],
 	'size': 0,
     }
+    self._procslimit = parlimit
+    if parlimit != None:
+      self._procspool = multiprocessing.Pool()
+      self._procslist = []
     self._buffer = stor_buffer()
     self._bufnum = 1
     self._bufsize = BLOCK_SIZE * 2
     self._checksum2 = None
     self._size = 0
+    self._something_written = False
     self._onlyifnew = onlyifnew
-    if compression == ('lzma', 10): self._compress = lambda buffer: lzma.compress(buffer, {'level': 9, 'extreme': True})
-    elif compression[0] == 'lzma': self._compress = lambda buffer: lzma.compress(buffer, {'level': compression[1]})
-    elif compression[0] == 'gzip': self._compress = lambda buffer: zlib.compress(buffer, compression[1])
-    elif compression[0] == 'bzip2': self._compress = lambda buffer: bz2.compress(buffer, compression[1])
-    elif compression[0] == None: self._compress = lambda buffer: buffer
-    else:
-      raise SystemError("Unsupported compression type %s" % repr(compression))
+    self._compression = compression
 
   def write(self, buffer):
     """Write function.
@@ -280,7 +356,7 @@ class stor_output(stor_stream):
     """Set the parent archive name."""
     self._metadata['parent'] = parent
 
-  def block_exist(self, buffer, begin, end, checksum1):
+  def _block_already_exists(self, buffer, begin, end, checksum1):
     """Checks if the given block already exists.
 
     Args:
@@ -306,12 +382,14 @@ class stor_output(stor_stream):
     """
     self.flush_buffer(BLOCK_SIZE)
     self.commit_buffer(self._buffer.pop(), 0)
-    if not self._onlyifnew or self._size > 0:
+    if not self._onlyifnew or self._something_written:
+      self._finish_write(0)
       self._metadata['usage'] = self._size
       self._metadata['size'] = self._pos
       buffer = pickle.dumps(self._metadata)
       self._bufnum = 0
       self.commit_buffer(buffer, 0, meta = True)
+      self._finish_write(0)
     stor_stream.close(self)
 
   def flush_buffer(self, tosize):
@@ -321,7 +399,7 @@ class stor_output(stor_stream):
     """
     while len(self._buffer) >= tosize:
       buffer = self._buffer.value()
-      found, pos, dest, fcks, cks = adler64.search(self._known_blocks, self.block_exist, buffer, BLOCK_SIZE)
+      found, pos, dest, fcks, cks = adler64.search(self._known_blocks, self._block_already_exists, buffer, BLOCK_SIZE)
       if found:
 	destlen = len(dest)
 	if destlen == BLOCK_SIZE: self.commit_buffer(dest, fcks)
@@ -334,6 +412,13 @@ class stor_output(stor_stream):
       else:
 	self.commit_buffer(self._buffer.pop(BLOCK_SIZE), fcks)
 
+  def _finish_write(self, limit):
+    while len(self._procslist) > limit:
+      proc = self._procslist.pop(0)
+      buffer, bufnum, checksum1, checksum2 = proc.get()
+      self._size += len(buffer)
+      self._backend.write_buffer(buffer, bufnum, checksum1, checksum2)
+
   def commit_buffer(self, buffer, checksum1int, meta = False):
     """Write data to the beckend.
 
@@ -348,23 +433,31 @@ class stor_output(stor_stream):
     if checksum1int == 0 and self._known_parts.has_key(checksum2):
       self._metadata['blocks'].append((self._known_parts[checksum2], checksum1, checksum2, buflen))
     else:
-      if meta: buffer = bz2.compress(buffer, 9)
-      else: buffer = self._compress(buffer)
-      bufrest = len(buffer) % 16
-      if bufrest > 0:
-	buffer = buffer + "\0" * (16 - bufrest)
-      if self._aes:
-	buffer = self._aes.encrypt(buffer)
-      self._size += len(buffer)
-      self._backend.write_buffer(buffer, self._bufnum, checksum1, checksum2)
+      self._something_written = True
+      msgencode_args = [
+          buffer,
+          self._aes,
+          meta,
+          self._compression,
+          self._bufnum,
+          checksum1,
+          checksum2,
+      ]
+      if self._procslimit != None:
+        self._procslist.append(self._procspool.apply_async(message_encode, msgencode_args))
+        self._finish_write(self._procslimit)
+      else:
+        buffer, bufnum, checksum1, checksum2 = apply(message_encode, msgencode_args)
+        self._size += len(buffer)
+        self._backend.write_buffer(buffer, bufnum, checksum1, checksum2)
       self._metadata['blocks'].append((self._bufnum, checksum1, checksum2, buflen))
-    if self._bufnum != 0: self._bufnum += 1
+    if not meta: self._bufnum += 1
     else: self._archive = checksum2
 
 class stor_input(stor_stream):
   """Storage input stream."""
 
-  def __init__(self, password, archive):
+  def __init__(self, password, archive, parlimit = 4):
     """Initialization of input stream.
 
     Args:
@@ -372,11 +465,14 @@ class stor_input(stor_stream):
       archive (str): Name of the archive (checksum).
     """
     stor_stream.__init__(self, password)
+    self._procslimit = parlimit
+    if parlimit != None:
+      self._procspool = multiprocessing.Pool()
+      self._procslist = []
     self._archive_checksum = archive
-    self._buffer = ""
-    self._bufferlen = 0
-    self._bufferpos = 0
+    self._buffer = stor_buffer(splitsize = 10240)
     self._archive = archive
+    self._prof = 0
 
   def open(self, backend):
     """Initialize backend for reading.
@@ -387,28 +483,76 @@ class stor_input(stor_stream):
     stor_stream.open(self, backend)
     if self._archive_checksum not in self._known_metadata:
       raise SystemError("Archive %s not found." % repr(self._archive_checksum))
-    self._metadata = self._backend.read_buffer(0, "%016X" % 0, self._archive_checksum,
-	lambda buf: pickle.loads(self._decode_message(buf, self._archive_checksum, 0, meta = True)))
-    self._decompress = lambda buffer: bz2.decompress(buffer)
-    if self._metadata['compression'][0] == 'lzma': self._decompress = lambda buffer: lzma.decompress(buffer)
-    elif self._metadata['compression'][0] == 'gzip': self._decompress = lambda buffer: zlib.decompress(buffer)
-    elif self._metadata['compression'][0] == 'bzip2': self._decompress = lambda buffer: bz2.decompress(buffer)
-    elif self._metadata['compression'][0] == None: self._decompress = lambda buffer: buffer
-    else: raise SystemError("Unsupported compression type %s" % repr(compression))
+    ck1 = "%016X" % 0
+    ck2 = self._archive_checksum
+    bufgen = self.read_buffer(0, ck1, ck2)
+    self._metadata = None
+    for buffer in bufgen:
+      try:
+        buffer, _, _, _, _ = message_decode(buffer, self._aes, self._pass, True, None, 0, ck1, ck2, 0)
+        self._metadata = pickle.loads(buffer)
+      except Exception, err: continue
+      break
+    if self._metadata == None:
+      raise SystemError('Failed to read metdata')
+    self._compression = self._metadata['compression']
     self._blocks = self._metadata['blocks'][:]
 
+  def bocks(self):
+    return [(0, "%016X" % 0, self._archive_checksum),] + sef._metadata['blocks']
+
+  def close(self):
+    stor_stream.close(self)
+
   def read(self, size):
     """Read data of given size.
 
     Use this function to read the data.
     """
-    while len(self._buffer) < size and len(self._blocks) > 0:
-      num, ck1, ck2, blen = self._blocks.pop(0)
-      data = self._backend.read_buffer(num, ck1, ck2, lambda buf: self._decode_message(buf, ck2, blen))
-      self._buffer = self._buffer + data
-      self._bufferlen += len(data)
-    data = self._buffer[:size]
-    self._buffer = self._buffer[len(data):]
+    data = None
+    if self._procslimit != None:
+      while len(self._buffer) < size and (len(self._blocks) > 0 or len(self._procslist) > 0):
+        data = None
+        while len(self._procslist) < self._procslimit and len(self._blocks) > 0:
+          num, ck1, ck2, blen = self._blocks.pop(0)
+          buffer_generator = self.read_buffer(num, ck1, ck2)
+          self._procslist.append((
+            buffer_generator,
+            self._procspool.apply_async(message_decode, [
+              buffer_generator.next(),
+              self._aes, self._pass, False, self._compression,
+              num, ck1, ck2, blen,
+            ]),
+            num, ck1, ck2, blen))
+        if self._procslimit != None:
+          bufgen, proc, num, ck1, ck2, blen = self._procslist.pop(0)
+          try: data, num, ck1, ck2, blen = proc.get()
+          except Exception, err: data = None
+          while data == None:
+            data = bufgen.next()
+            try:
+              data, num, ck1, ck2, blen = apply(message_decode, [
+                data, self._aes, self._pass, False, self._compression,
+                num, ck1, ck2, blen,
+              ])
+            except Exception, err: data = None
+        self._buffer.push(data)
+    else:
+      while len(self._buffer) < size and len(self._blocks) > 0:
+        data = None
+        num, ck1, ck2, blen = self._blocks.pop(0)
+        for buffer in self.read_buffer(num, ck1, ck2):
+          try:
+            data, num, ck1, ck2, blen = apply(message_decode, [
+              data, self._aes, self._pass, False, self._compression,
+              num, ck1, ck2, blen,
+            ])
+          except Exception, err: continue
+          break
+        self._buffer.push(data)
+    x = time.time()
+    data = self._buffer.pop(size)
+    self._prof += time.time() - x
     return data
 
   def metainfo(self):
@@ -431,11 +575,18 @@ class stor_util(stor_stream):
     stor_stream.open(self, backend)
     checksum1 = "%016X" % 0
     for metadata in self._known_metadata:
-      try: data = self._backend.read_buffer(0, checksum1, metadata,
-            lambda buf: pickle.loads(self._decode_message(buf, metadata, 0, meta = True)))
+      bufgen = self.read_buffer(0, checksum1, metadata)
+      data = None
+      try:
+        for buffer in bufgen:
+          try:
+            data, _, _, _, _ = message_decode(buffer, self._aes, self._pass, True, None, 0, checksum1, metadata, 0)
+            data = pickle.loads(data)
+          except Exception, err: continue
+          break
       except Exception, err:
-	self._errors.append((metadata, err))
-	continue
+        self._errors.append((metadata, err))
+        continue
       self._metadata_data[metadata] = data
 
   def search(self, keywords):
@@ -492,4 +643,26 @@ class stor_util(stor_stream):
 	keywords[kwd] = keywords.get(kwd, 0) + 1
     return (keywords, self._errors)
 
+  def block_exists(self, number, checksum1, checksum2):
+    if type(checksum1) != type(0):
+      checksum1 = int(checksum1, 16)
+    if number == 0:
+      if checksum2 in self._known_metadata:
+        return True
+    elif checksum1 == 0:
+      if self._known_parts.has_key(checksum2):
+        if number in self._known_parts[checksum2]:
+          return True
+    else:
+      if self._known_blocks.has_key(checksum1):
+        if self._known_blocks[checksum1].has_key(checksum2):
+          if number in self._known_blocks[checksum1][checksum2]:
+            return True
+    return False
+
+  def blocks(self, archive):
+    ck1 = "%016X" % 0
+    metadata = None
+    return [(0, ck1, archive, 0)] + self._metadata_data[archive]['blocks'][:]
+
 __all__ = ('stor_output', 'stor_input', 'stor_util')
